import java.io.*;
import java.net.*;

public class DebugTCPThread implements Runnable {

	private int port = 0;
	private Thread netThread = null;
	private Thread receiveThread = null;
	private DebugTCPUI userInterface = null;
	private DebugTCPState st;
	private ServerSocket serverSocket = null;
	private Socket clientSocket = null;
	private boolean operationInterrupted = false;
	private byte[] outputData;
	private DataOutputStream out;
	private ReceiveData receiveData = null;
	private String remoteHost;

	public DebugTCPThread(DebugTCPUI ui, DebugTCPState stParam) {
		userInterface = ui;
		st = stParam;
	}

	public void startListening(int portParam) {
		if (serverSocket == null) {
			port = portParam;
			netThread = new Thread(this,"NetworkThread");
			netThread.start();
		}
	}

	public void connectTo(String remoteHostParam, int remotePortParam) {
		if (clientSocket == null) {
			port = remotePortParam;
			remoteHost = remoteHostParam;
			netThread = new Thread(this,"NetworkThread");
			netThread.start();
		}
	}

	public void run() {
		if (st.actingAsServer) {
			// Running as server. Listen on a port.
			operationInterrupted = false; // protection against inconsistent state
			try {
				serverSocket = new ServerSocket(port,1);
				st.isListening = true;
				userInterface.updateConnectionState();
				while(st.isListening) {
					try {
						clientSocket = serverSocket.accept();
						initializeConnection();
						waitForOutputData();
					} catch (IOException ex) {
						if (operationInterrupted) {
							operationInterrupted = false;
						} else {
							st.isListening = false;
							String msg = ex.getMessage();
							userInterface.popupErrorMessage("Accept failed: " + msg);
						}
					} finally {
						if(clientSocket != null) {
							try {
								clientSocket.close();
							} catch (IOException ex) {
								userInterface.popupErrorMessage(
										"Cannot close client socket.");
							} finally {
								clientSocket = null;
								userInterface.updateConnectionState();
							}
						}
					}
				}
			} catch (IOException ex) {
				String msg = ex.getMessage();
				userInterface.popupErrorMessage("Cannot create ServerSocket: " + msg);
			} catch (IllegalArgumentException ex) {
				String msg = ex.getMessage();
				userInterface.popupErrorMessage("Illegal port: " + msg);
			} finally {
				serverSocket = null;
				userInterface.updateConnectionState();
			}
		} else {
			// Running as client. Connect to another host.
			try {
				clientSocket = new Socket();
				InetSocketAddress sockAddr = new InetSocketAddress(remoteHost, port);
				clientSocket.connect(sockAddr,15000); // timeout is 15 seconds
				st.isAttemptingToConnect = false;
				initializeConnection();
				waitForOutputData();
			} catch (UnknownHostException ex) {
				userInterface.popupErrorMessage("Unknown host: " + ex.getMessage());
			} catch (IOException ex) {
				userInterface.popupErrorMessage("IOException: " + ex.getMessage());
			} finally {
				st.isAttemptingToConnect = false;
				userInterface.updateConnectionState();
				if(clientSocket != null) {
					try {
						clientSocket.close();
					} catch (IOException ex) {
						userInterface.popupErrorMessage(
								"Cannot close client socket.");
					} finally {
						clientSocket = null;
					}
				}
			}
		}
	}

	public void stopListening() {
		st.isListening = false;
		operationInterrupted = true;
		try {
			serverSocket.close();
		} catch (IOException ex) {
			userInterface.popupErrorMessage("Cannot close server socket.");
		}
	}

	public synchronized void closeConnection() {
		st.isConnected = false;
		notify();
	}

	private void initializeConnection() {
		try {
			out = new DataOutputStream(clientSocket.getOutputStream());
			st.isConnected = true;
			st.isNewConnection = true;
			userInterface.updateConnectionState();
			receiveData = new ReceiveData();
			receiveThread = new Thread(receiveData,"ReceiveThread");
			receiveThread.start();

			InetAddress addr = clientSocket.getInetAddress();
			userInterface.displayRemoteAddressAndPort(
					addr.getHostAddress(), clientSocket.getPort());

		} catch (IOException ex) {
			userInterface.popupErrorMessage("IOException in initializeConnection(): "
					+ ex.getMessage());
		}
	}

	private synchronized void waitForOutputData() {
		while(st.isConnected) {
			try {
				wait();
				if(st.isConnected) {
					out.write(outputData);
				}
			} catch (InterruptedException ex) {
				userInterface.popupErrorMessage("InterruptedException in waitForOutputData(): "
						+ ex.getMessage());
			} catch (IOException ex) {
				st.isConnected = false;
				userInterface.popupErrorMessage("IOException in waitForOutputData(): "
						+ ex.getMessage());
			}
		}
	}

	public synchronized void sendData(byte[] data, int dataLength) {
		outputData = new byte[dataLength];
		System.arraycopy(data,0,outputData,0,dataLength);
		notify();
	}

	public class ReceiveData implements Runnable {
		private DataInputStream in = null;
		public void run() {
			try {
				int rv;
				byte[] buf = new byte[512];
				in = new DataInputStream(clientSocket.getInputStream());
				while(st.isConnected) {
					rv = in.read(buf);
					if(rv < 0) {
						closeConnection();
					} else {
						userInterface.displayReceivedData(buf,rv);
					}
				}
			} catch (IOException ex) {
				if (st.isConnected) { closeConnection(); }
			}
		}
	}

}
